{-# OPTIONS_GHC -fallow-undecidable-instances #-}
-- Just needed for the show operator.
module AST.Prolog where

import Data.List (nub, intersperse, concat)

-- Data structures representing a Prolog program.
-- They are named in a generic way, so I recommend using this module
-- as import qualified.

type Const   = [Char]

-- egallego@babel: 23 Sep 2004, 18:22
-- Monadic implementation of Term data type has been delayed.
-- For now we use a non monadic data type and manual traversals.

-- The general Term data type.
data Term a = TVar a
            | TStruct Const [Term a]
              deriving (Eq, Ord, Show)

-- Needs thinking about general traversal of an AST.
-- f is seen as a renaming function.

instance Functor Term where
    fmap f (TVar v)       = TVar $ f v
    fmap f (TStruct n lt) = TStruct n $ map (fmap f) lt

instance Monad Term where
    return x              = TVar x
    (TVar v)       >>= f = f v
    (TStruct n lt) >>= f = TStruct n (map (>>= f) lt)

-- Basic prolog structures, enforces basic sanity checks, such as
-- variables a preds.
data Atom a = Atom Const [Term a]
            deriving (Eq, Show)

-- This lacks direct support for negation.
data Clause a = Clause (Atom a) [Atom a]
              deriving (Eq, Show)

-- Name and arity
type Predicate = (Const, Int)

-- STerm := Prolog term.
type STerm = Term String

-- ITerm := Relational terms (normalized to ints.
type ITerm = Term Int

-- Auxiliary functions.
arity :: Atom a -> Int
arity (Atom _ l) = length l

--
term_arity :: Term a -> Int
term_arity (TVar _)      = 0
term_arity (TStruct _ l) = length l

-- Constructors and deconstructors.
functor :: Term a -> Const
functor (TVar a)      = error "Trying to access function symbol for a variable"
functor (TStruct f _) = f

args :: Term a -> [Term a]
args (TVar a)      = error "Trying to access arguments for a variable"
args (TStruct _ a) = a

-- Get the normal name for a predicate (fixme: modules)
getp_rule :: Clause a -> Predicate
getp_rule (Clause a _) = getp_atom a

getp_atom :: Atom a -> Predicate
getp_atom (Atom n l) = (n, length l)

-- Variable extraction.
vars_clause :: Eq a => Clause a -> [a]
vars_clause (Clause a ll) =
             let va = vars_atom a
                 vl = nub $ concat $ map vars_atom ll in
                      nub $ (va ++ vl)

vars_atom :: Eq a => Atom a -> [a]
vars_atom (Atom _ lt) = nub $ concat $ map vars_term lt

vars_term :: Eq a => Term a -> [a]
vars_term (TVar v) = [v]
vars_term (TStruct _ lt) = nub $ concat $ map vars_term lt

vars_list :: Eq a => [Term a] -> [a]
vars_list l = nub . concat $ map vars_term l

--
is_var :: Term a -> Bool
is_var (TVar _) = True
is_var _         = False


-- instance Show a => Show (Term a) where
--     show (CTVar x)      = show x
--     show (CTStruct h a) = if length a == 0 then
--                               h
--                           else h ++ "(" ++ (concat $ intersperse "," $ map show a) ++ ")"
data TagType = UnTagged
             | Freeze
             | Rewrite
               deriving (Eq, Show)

data Tag ty = CTag TagType ty
              deriving (Eq, Show)

instance Functor Tag where
    fmap f (CTag m ty) = CTag m (f ty)

instance Monad Tag where
    return x = CTag UnTagged x
    CTag m ty >>= f = f ty

change_tag :: TagType -> Tag a -> Tag a
change_tag nt (CTag _ e) = CTag nt e

tagList :: [a] -> Tag [ Tag a ]
tagList l = return $ fmap return l
