-----------------------------------------------------------------
-- |
-- Module      : Engine.Unification.MM
-- Copyright   : (c) Babel Group 2007
-- License     : GPL
--
-- Maintainer  : egallego@babel.ls.fi.upm.es
-- Stability   : experimental
-- Portability : portable
--
-- Martelli Montanary unification algorithm.
--
-----------------------------------------------------------------

module Engine.Unification.MM where
-- Fix the module once it is stable.
-- module Engine.Unification.MM(unify) where

-- Martelli Montanary unification

import AST.Prolog
import Translator.String.Prolog

import Data.Maybe

import Data.Set (Set)
import qualified Data.Set as Set

import Data.List

import Debug.Trace

tt s = let Right t = trans_term s in t

-- Note, MM could be implemented in a much more efficient way now that
-- I know it in depth. Speed should be comparable to the Pascal
-- version. For one, equations should be indexed by the left side
-- vars, and so on. Would make an interesting undergraduate project
-- (including a stricness analysis on the code)

-- The Rules of Optimization are simple. Rule 1: Don't do it. Rule 2
-- (for experts only): Don't do it yet. Michael A. Jackson, 1975,
-- Principles of Program Design.

-- Topological view on unification. (TODO)

-- |
-- unify :: String -> String -> (Unifier, Unifier)
-- unify s1 s2 = let (Right t1, Right t2) = (trans_term s1, trans_term s2) in
--               unifymm t1 t2

type MultiSet a = [a]
type MultiEq = [(Set STerm, Set STerm)]

-- Example terms. (Should be used with HUnit)
m1 = [tt "f(X1, g(a,f(X5,b)))", tt "f(h(c), g(X2, f(b,X5)))", tt "f(h(X4), g(X6, X3))"]
m2 = [tt "h(X3, h(X2, X2))", tt "h(h(h(X1,X1), X2), X3)", tt "X4"]
m3 = [tt "f(X1, g(X2,X3), X2, b)", tt "f(g(h(a,X5), X2), X1, h(a, X4), X4)"]
m4 = [tt "X", tt "f(Y)", tt "f(f(Z))", tt "f(f(g(k)))", tt "f(f(g(m)))"]
-- From patweg
m5 = [tt "a(b(V), c(U,V))", tt "a(b(W),c(W, d(X,Y)))"]

mgumm :: MultiSet STerm -> Maybe MultiEq
mgumm ms =
    do unif <- unifymm ms
       let unif' = reverse unif
           res   = mgumm' unif
       return res

mgumm' [] = []
mgumm' ((x@(_,t)):xs) =
    if Set.null t then
        x:(mgumm' xs)
    else
        (x: (mgumm' ((substitute x) xs)))

-- | MM algorithm
--   Keep in mind that this returns the triangular multiequation, not the MGU.
unifymm :: MultiSet STerm -> Maybe MultiEq
unifymm ms =
    do let mmf = (TVar "_x"):ms -- This is avoidable by solving unifyloop here.
           vs  = nub $ concat $ map vars_term ms
           emptyeq = [] -- map (\v -> (Set.singleton (CTVar v), Set.empty)) vs
       (_, me) <- splitTerms mmf
       unifyloop (emptyeq ++ me) []

-- Linear in the order of new added equations. O(n_eqs)
unifyloop :: MultiEq -> MultiEq -> Maybe MultiEq
unifyloop [] t = return t
unifyloop u  t = do
  (x, z') <- selecteq u -- Will fail if cycle.
  case emptyEq x of
    True  -> do unifyloop z' (t++[x]) -- Hack due to incorrect ordering from selecteq.
                                      -- But we have to deal with left remaining eqs anyways.
    False -> do
      u' <- mered x
      let (sol:u'') = u'
          z'' = z' ++ u''
          ured = compacteq z''
          neweq = substitute sol ured
      unifyloop neweq (sol:t)

-- O(n_subs*term_size)
substitute :: (Set STerm, Set STerm) -> MultiEq -> MultiEq
substitute (v, t) me =
    let toVar (TVar v) = v
        vars = map toVar $ Set.toList v
        [st] = Set.toList t
        subs = map (flip subVar st) vars in
    foldr (\s m -> s m) me (map sub subs)

sub :: Substitution -> [(Set STerm, Set STerm)] -> [(Set STerm, Set STerm)]
sub f = map (\(v,t) -> (v, Set.map (apply f) t))

-- Here will go cycle detection and ordering.
--
selecteq :: MultiEq -> Maybe ((Set STerm, Set STerm), MultiEq)
selecteq = toposelect
--selecteq (x:xs) = return h

-- This deserves some commentary, I know.
toposelect :: MultiEq -> Maybe ((Set STerm, Set STerm), MultiEq)
toposelect eqs =
    let rhvar = foldr getrhsvars Set.empty eqs
        (cands, nc) = partition (notLHS rhvar) eqs in
    case cands of
      [] -> Nothing
      _  -> return (head cands, tail cands ++ nc )

emptyEq (v, t) = Set.null t

getrhsvars :: (Set STerm, Set STerm) -> Set STerm -> Set STerm
getrhsvars (_, rhs) vset = Set.union rhvars vset
    where rhvars = vars_set rhs

vars_set :: Set STerm -> Set STerm
vars_set s = let vl = concat $ map vars_term (Set.toList s) in
                 Set.fromList $ map TVar vl

-- Returns true if
notLHS :: Set STerm -> (Set STerm, Set STerm) -> Bool
notLHS vs (lv, _) = Set.null $ Set.intersection vs lv

-- I'd guess this one is O(n_eqs^2) due to split [x1, x2 ... x_n]
-- Sure we could do better.
-- Compacts equations which have some element in common in the left side.
compacteq :: MultiEq -> MultiEq
compacteq ms =
    join sp
    where sp = split ms
          split :: MultiEq -> [MultiEq]
          split []   = []
          split l@((v,t):xs) = let (pos, neg) = partition (cond v) l in
                               pos : (split neg)

          cond vs1 (vs2, _) = (not . Set.null) $ Set.intersection vs1 vs2

          join l = map (joinms . unzip) l
          joinms (v,t) =  (Set.unions v, Set.unions t)

-- multieq reduction.
mered :: (Set STerm, Set STerm) -> Maybe MultiEq
mered (s, m) = do
  (c,f) <- splitTerms $ Set.toList m -- commmon, frontier
  return ((s, Set.singleton c):f)

-- O(n)
-- | Returns (Common part, Frontier)
splitTerms :: MultiSet STerm -> Maybe (STerm, MultiEq)
splitTerms ms =
    case find is_var ms of -- Full O(n)
      Just var -> return (var, makemulteq ms)
      Nothing  ->
          do h <- check_common_head ms
             let (TStruct fh arg) = h
             if arg == []
              then do
                return (h, [])
              else do
                let ms' = transpose $ map args ms
                x <- mapM splitTerms ms'
                let (cp, fr) = unzip x
                    (TStruct fh _):_ = ms
                return (TStruct fh cp, concat fr)

-- Checks common heads and returns the functor
check_common_head :: MultiSet STerm -> Maybe STerm
check_common_head (s@(TStruct f ar):xs) =
    let l = length ar in
    if all (\t -> functor t == f ) xs &&
       all (\t -> (length.args) t == l) xs
    then
        Just s
    else
        Nothing

-- We should kill this.
makemulteq :: MultiSet STerm -> MultiEq
makemulteq ms = [(Set.fromList $ vars, Set.fromList $ terms)]
                 where (vars, terms) = partition is_var ms

-- | Non deterministic algorithm. Without occurs check. Able to
-- partially handle infinite terms.
type Eqt      = (STerm, STerm)
type EqSet   = [Eqt]
type Unifier = EqSet

unifynd :: String -> String -> (Unifier, Unifier)
unifynd s1 s2 = let (Right t1, Right t2) = (trans_term s1, trans_term s2) in
              unify1 t1 t2
unify1 :: STerm -> STerm -> (Unifier, Unifier)
unify1 t1 t2 = reduceEqSet ([(t1,t2)], [])

reduceEqSet :: (EqSet, EqSet) -> (EqSet, EqSet)
reduceEqSet ([], sol) = ([], sol)
reduceEqSet (eqs, sol) =
    let (eq, eqs')  = trace (show eqs) selectEq eqs
        req         = reduceEq eq
    in case req of
         Kill    -> reduceEqSet (eqs', sol)
         New e   -> reduceEqSet (mergeEqSet e eqs', sol)
         VarElim s ->
             let sub = apply s in
             reduceEqSet (map (\(v,t) -> (sub v, sub t)) eqs', eq:sol)

selectEq (x:xs) = (x,xs)

data Action = Kill
            | New EqSet
            | VarElim Substitution

reduceEq :: Eqt -> Action
reduceEq (s@(TStruct _ _), v@(TVar _)) = New [(v, s)]
reduceEq e@(TVar t1, TVar t2) = if t1 == t2 then Kill else VarElim $ subVar t1 (TVar t2)
reduceEq (TStruct s1 l1, TStruct s2 l2) =
    if s1 == s2 &&
       length l1 == length l2 then -- MM papers doesn't account bad typed terms.
        New $ zip l1 l2
    else error "No"
reduceEq (TVar v, t) = VarElim $ subVar v t

mergeEqSet :: EqSet -> EqSet -> EqSet
mergeEqSet = (++)

-- Copied from old unify.
type Substitution = String -> STerm

apply                   :: Substitution -> STerm -> STerm
apply s (TVar v)       = s v
apply s (TStruct n lt) = TStruct n (map (apply s) lt)

subVar :: String -> STerm -> Substitution
subVar v t = \x -> if v == x then t else (TVar x)

