-----------------------------------------------------------------
-- |
-- Module      : Engine.Unification.MM2
-- Copyright   : (c) Babel Group 2007
-- License     : GPL
--
-- Maintainer  : egallego@babel.ls.fi.upm.es
-- Stability   : experimental
-- Portability : portable
--
-- Martelli Montanary unification algorithm.
--
-----------------------------------------------------------------

module Engine.Unification.MM2 where
-- Fix the module once it is stable.
-- module Engine.Unification.MM(unify) where

-- Martelli Montanary unification

import AST.Prolog
import Translator.String.Prolog

import Data.Maybe

import Data.Set (Set)
import qualified Data.Set as Set

import Data.List

import Debug.Trace

tt s = let Right t = trans_term s in t

type Var = String
type MultiSet a = [a]
-- Contains {vars, terms}
type MultiEq = (Set Var, Set STerm)

-- Example terms. (Should be used with HUnit)
m1 = [tt "f(X1, g(a,f(X5,b)))", tt "f(h(c), g(X2, f(b,X5)))", tt "f(h(X4), g(X6, X3))"]
m2 = [tt "h(X3, h(X2, X2))", tt "h(h(h(X1,X1), X2), X3)", tt "X4"]
m3 = [tt "f(X1, g(X2,X3), X2, b)", tt "f(g(h(a,X5), X2), X1, h(a, X4), X4)"]
m4 = [tt "X", tt "f(Y)", tt "f(f(Z))", tt "f(f(g(k)))", tt "f(f(g(m)))"]
-- From patweg
m5 = [tt "a(b(V), c(U,V))", tt "a(b(W),c(W, d(X,Y)))"]

mgumm2 :: MultiSet STerm -> Maybe [MultiEq]
mgumm2 ms =
    do unif <- unifymm ms
       let unif' = reverse unif
           res   = mgumm' unif
       return res

mgumm' :: [MultiEq] -> [MultiEq]
mgumm' [] = []
mgumm' ((x@(_,t)):xs) =
    if Set.null t then
        x:(mgumm' xs)
    else
        (x: (mgumm' ((substitute x) xs)))

-- | MM algorithm
--   Keep in mind that this returns the triangular multiequation, not the MGU.
unifymm :: MultiSet STerm -> Maybe [MultiEq]
unifymm ms =
    unifyloop [((Set.fromList ["_x"]), Set.fromList ms)] []

-- Linear in the order of new added equations. O(n_eqs)
unifyloop :: [MultiEq] -> [MultiEq] -> Maybe [MultiEq]
unifyloop [] t = return t
unifyloop u  t = do
  (x, z') <- selecteq u -- Will fail if cycle.
  u' <- mered x
  let (sol:u'') = u' -- As mred returns the solution
      z'' = z' ++ u''
      ured = compacteq z''
      neweq = substitute sol ured
  unifyloop neweq (sol:t)

-- O(n_subs*term_size)
substitute :: MultiEq -> [MultiEq] -> [MultiEq]
substitute (v, t) me =
    let [st] = Set.toList t
        subs = map (flip subVar st) (Set.toList v) in
    foldr (\s m -> s m) me (map sub subs)

sub :: Substitution -> [(Set Var, Set STerm)] -> [(Set Var, Set STerm)]
sub f = map (\(v,t) -> (v, Set.map (apply f) t))

-- Here will go cycle detection and ordering.
--
selecteq :: [MultiEq] -> Maybe (MultiEq, [MultiEq])
selecteq = toposelect
--selecteq (x:xs) = return h

-- This deserves some commentary, I know.
toposelect :: [MultiEq] -> Maybe (MultiEq, [MultiEq])
toposelect eqs =
    let rhvars = map getrhsvars eqs
        rvset  = foldr Set.union Set.empty rhvars
        (cands, nc) = partition (notLHS rvset) eqs in
    case cands of
      []     -> Nothing
      (x:xs) -> return (x, xs ++ nc ) -- Pick the first free multieq.

getrhsvars :: MultiEq -> Set Var
getrhsvars (_, rhs) = vars_set rhs

vars_set :: Set STerm -> Set Var
vars_set s = set_concat $ Set.map (\x -> Set.fromList (vars_term x)) s

-- Looks like a good addition to the standard lib.
set_concat :: Ord a => Set (Set a) -> Set a
set_concat s = Set.fold Set.union Set.empty s

-- Returns true if
notLHS :: Set Var -> MultiEq -> Bool
notLHS vs (lv, _) = Set.null $ Set.intersection vs lv

-- I'd guess this one is O(n_eqs^2) due to split [x1, x2 ... x_n]
-- Sure we could do better.
-- Compacts equations which have some element in common in the left side.
compacteq :: [MultiEq] -> [MultiEq]
compacteq ms =
    join sp
    where sp = split ms
          split :: [MultiEq] -> [[MultiEq]]
          split []   = []
          split l@((v, t):xs) = let (pos, neg) = partition (cond v) l in
                                pos : (split neg)

          cond vs1 (vs2, _) = (not . Set.null) $ Set.intersection vs1 vs2

          join l = map (joinms . unzip) l
          joinms (v,t) =  (Set.unions v, Set.unions t)

-- multieq reduction.
mered :: MultiEq -> Maybe [MultiEq]
mered (vars, terms) = do
  (c, f) <- splitTerms terms -- commmon, frontier
  return $ (vars, Set.singleton c) : f

-- O(n)
-- | Returns (Common part, Frontier)
-- Fixme: It has a bug whem ms is empty.
splitTerms :: Set STerm -> Maybe (STerm, [MultiEq])
splitTerms ms =
    let v = Set.filter is_var ms in -- Full O(n)
    if Set.null v then
        do h <- check_common_head (Set.toList ms) -- fix to use sets.
           let (TStruct fh arg) = h
           if arg == []
            then do
               return (h, [])
            else do
               let ms' = transpose $ map args (Set.toList ms)
               x <- mapM (\x-> splitTerms (Set.fromList x)) ms'
               let (cp, fr) = unzip x
               return (TStruct fh cp, concat fr)
    else return (Set.findMin v, makemulteq (Set.toList ms))

-- Checks common heads and returns the functor
check_common_head :: MultiSet STerm -> Maybe STerm
check_common_head (s@(TStruct f ar):xs) =
    let l = length ar in
    if all (\t -> functor t == f ) xs &&
       all (\t -> (length.args) t == l) xs
    then
        Just s
    else
        Nothing

-- We should kill this.
makemulteq :: MultiSet STerm -> [MultiEq]
makemulteq ms = [(Set.fromList $ (map (\(TVar x) -> x) vars), Set.fromList $ terms)]
                 where (vars, terms) = partition is_var ms

-- Copied from old unify.
type Substitution = String -> STerm

apply                   :: Substitution -> STerm -> STerm
apply s (TVar v)       = s v
apply s (TStruct n lt) = TStruct n (map (apply s) lt)

subVar :: String -> STerm -> Substitution
subVar v t = \x -> if v == x then t else (TVar x)

