{

-- This module implements the parsing of a Prolog grammar.

module Translator.String.Prolog.Syntax where

import AST.Prolog

import Translator.String.Prolog.Lexer

import Compiler.Module (Module, addDefinition)

import Data.Either
import Control.Monad.Error
import Control.Monad.State
import Control.Monad.Writer

}

%name prolog_parser program
%name clause_parser rule
%name term_parser   term

%tokentype { PrologToken }

%monad { PrologParseMonad } --{ (>>=) } { returnM }

-- Tokens of Prolog.
%token
      '.'             { PTSym "."  }
      ','             { PTSym ","  }
      ':-'            { PTSym ":-" }
      '('             { PTSym "("  }
      ')'             { PTSym ")"  }
      '['             { PTSym "["  }
      ']'             { PTSym "]"  }
      '|'             { PTSym "|"  }
      VAR             { PTVar $$   }
      CONST           { PTConst $$ }

%%
-- Grammar of EONF descriptions.

program :: { [Predicate] }
    : clause                     { [$1] }
    | program clause             { $1 ++ [$2] }

clause  :: { Predicate }
    : rule                       { % addClause (getp_rule $1) $1 }
    | atom '.'                   { % addClause (getp_atom $1) (Clause $1 []) }

rule    :: { Clause String}
    : atom ':-' rule_body '.'    { Clause $1 $3 }
    | ':-' rule_body '.'         { Clause (mkQuery $2) $2 }

rule_body :: { [Atom String] }
    : literal                    { [$1] }
    | rule_body ',' literal      { $1 ++ [$3] }


literal :: { Atom String }
    : atom                       { $1 }

atom    :: { Atom String }
    : CONST '(' term_list ')'    { Atom $1 $3 }
    | CONST                      { Atom $1 [] }

term_list :: { [Term String] }
    : term                       { [$1] }
    | term_list ',' term         { $1 ++ [$3] }

-- This is unused. Why?
generic_list :: { [a] }
    : term                       { [$1] }
    | term_list ',' term         { $1 ++ [$3] }

term    :: { Term String }
    : VAR                        { TVar    $1 }
    | struct                     { $1 }
    | list                       { $1 }

struct  :: { Term String }
    : CONST                      { TStruct $1 [] }
    | CONST '(' term_list ')'    { TStruct $1 $3 }

-- FIXME: Should support list like [1,2,3] that are the equiv of [1|[2|[3|[]]]]
list:: { Term String }
    : '[' ']'                    { TStruct "_btlnnil"  []}
    | '[' term ']'               { TStruct "_btlnlist" [$2, TStruct "_btlnnil" []] }
    | '[' term '|' term ']'      { TStruct "_btlnlist" [$2, $4] }


{

-------------------------
-- | Parser State monads.
--

type ParserState = Module Predicate (Clause String)

type PrologParseMonad a = ErrorT [Char] (State ParserState) a
-- type PrologParseMonad a = WriterT [Char] (State ParserState) a

-- FIXED: We should use monad transformers.
--returnM :: a -> PrologParseMonad a
--returnM a = return $ return a

--thenM :: PrologParseMonad a -> (a -> PrologParseMonad b) -> PrologParseMonad b
--thenM e f = case e of
--                 Right a -> f
--                 Left  b -> Left b

-- Predicate query
mkQuery :: Eq a => [Atom a] -> Atom a
mkQuery l =
    let d  = Clause (Atom "query" []) l
        nv = vars_clause d
        lt = map (\x -> TVar x) nv in
             Atom "query" lt

-- Adds a clause to a predicate.

addClause :: Predicate -> Clause String -> PrologParseMonad Predicate
addClause p c = lift $ do mod <- get
                          put (addDefinition mod p c)
                          return p

happyError :: [PrologToken] -> PrologParseMonad a
happyError _ = fail "Compile error"
--happyError t =
--    do tell $ "Compile error " ++ (show t)
--       throwError "Lanzando"

}
