module Util.StrictMap where

-- This implements a data structure with Strict 1 <-> 1 mapping.
-- It's used in eliminating x2 = a, x3 = a in prolog.

import qualified Data.Map as Map

data StrictMap a b = SM (Map.Map a b) (Map.Map b a)

reverseSM :: StrictMap a b -> StrictMap b a
reverseSM (SM fm1 fm2) = SM fm2 fm1

emptySM :: StrictMap a b 
emptySM = SM Map.empty Map.empty


addToSM   :: (Ord a, Ord b) => StrictMap a b -> a -> b -> StrictMap a b
addToSM sm@(SM fm1 fm2) o1 o2 = 
    let fm1'    = Map.insert o1 o2 fm1
	fm2'    = Map.insert o2 o1 fm2 in
	   SM fm1' fm2'
	

delFromFirstSM  :: (Ord a, Ord b) => StrictMap a b -> a -> StrictMap a b
delFromFirstSM sm@(SM fm1 fm2) o =
    case Map.lookup o fm1 of
         Nothing -> sm
	 Just o' -> let fm1'  = Map.delete o  fm1 
			fm2'  = Map.delete o' fm2 in
			   SM fm1' fm2'


delFromSecondSM :: (Ord a, Ord b) => StrictMap a b -> b -> StrictMap a b
delFromSecondSM sm@(SM fm1 fm2) o = 
    let sm' = reverseSM sm in
	  reverseSM $ delFromFirstSM sm' o

lookupFirstSM   :: (Ord a, Ord b) => StrictMap a b -> a -> Maybe b
lookupFirstSM sm@(SM fm1 fm2) o =
    Map.lookup o fm1 

lookupSecondSM  :: (Ord a, Ord b) => StrictMap a b -> b -> Maybe a
lookupSecondSM sm@(SM fm1 fm2) o =
    Map.lookup o fm2



-- Un StrictMap solo puede mapearse con una funcin si es del mismo tipo.
-- Por eso hacemos un map de dos funciones.
mapSM :: (Ord a, Ord b) => StrictMap a b -> (a -> b -> b) -> (b -> a -> a) -> StrictMap a b
mapSM sm@(SM fm1 fm2) f1 f2 = 
    let fm1' = Map.mapWithKey f1 fm1
	fm2' = Map.mapWithKey f2 fm2 in
	   SM fm1' fm2'

smToList        :: StrictMap a b -> [(a,b)]
smToList (SM fm1 _) = Map.toList fm1


listToSM :: (Ord a, Ord b) => [(a, b)] -> StrictMap a b
listToSM l =
    let fm1 = Map.fromList l
	fm2 = Map.fromList $ map (\(a,b) -> (b,a)) l in
	      SM fm1 fm2
